/*
 * host.c - host for multiap network ubus interface
 *
 * Copyright (C) 2021 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdbool.h>
#include <pthread.h>
#include <json-c/json.h>
#include <libubox/blobmsg.h>

#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <net/if.h>
#include <uci.h>

#include <easy/easy.h>
#include "host.h"
#include "host_config.h"
#include "debug.h"
#include "config.h"
#include "json_utils.h"

extern const char *ubus_socket;

struct u_list {
	char ifname[10][128];
	uint32_t count;
};

void remove_newline(char *buf)
{
	int len;

	len = strlen(buf) - 1;
	if (buf[len] == '\n')
		buf[len] = 0;
}

char *topology_stringify_address_types(enum address_type type)
{
	switch (type) {
	case HOST_ADDR_STATIC:
		return "Static";
	case HOST_ADDR_DHCP:
		return "DHCP";
	}
	return "";
}

char *topology_stringify_types(enum host_types type)
{
	switch (type) {
	case HOST_TYPE_ETHER:
		return "Ethernet";
	case HOST_TYPE_GHN:
		return "G.hn";
	case HOST_TYPE_HPNA:
		return "HPNA";
	case HOST_TYPE_HOMEPLUG:
		return "HomePlug";
	case HOST_TYPE_MOCA:
		return "MoCA";
	case HOST_TYPE_UPA:
		return "UPA";
	case HOST_TYPE_WIFI:
		return "Wi-Fi";
	case HOST_TYPE_OTHER:
		return "Other";
	}
	return "";
}

static void uobjx_lookup(struct ubus_context *ctx, struct ubus_object_data *obj, void *priv)
{
	const char *token = "wifi.ap";
	char *p, *mac;
	struct u_list *str = (struct u_list *)priv;

	p = strstr(obj->path, token);
	if (p != NULL) {
		strcpy(str->ifname[str->count], obj->path);
		str->count = str->count + 1;
	}
fail:
	return;
}

static int host_get_wifi_nodes(struct ubus_context *ctx, struct u_list *c)
{
	c->count = 0;
	int i = 0;

	ubus_lookup(ctx, NULL, uobjx_lookup, c);
	for (i = 0; i < c->count; i++)
		dbg("%d %s\n", __LINE__, c->ifname[i]);
	return 0;
}

int host_nodes(struct ubus_context *ctx,
		struct ubus_object *obj,
		struct ubus_request_data *req,
		const char *method,
		struct blob_attr *msg)
{
	struct blob_buf bb;
	struct topologyd_private *priv =
		container_of(obj, struct topologyd_private, obj);

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	host_dump_node_nbr(priv, &bb);

	ubus_send_reply(ctx, req, bb.head);
	blob_buf_free(&bb);

	return 0;
}

void host_dump_node_nbr(struct topologyd_private *priv, struct blob_buf *b)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	struct uci_section *section;
	char mac_str[18] = { 0x0 };
	void *nodes_array;
	struct host_node *p;
	int i;
	void *table1 = NULL;
	struct tm *info;
	time_t tmp_t;
	char str_tm[32] = {0};
	int j = 0;

	ctx = uci_alloc_context();
	if (!ctx)
		return;

	if (uci_load(ctx, "hosts", &pkg)) {
		uci_free_context(ctx);
		return;
	}

	dbg("Inside %s %d\n", __func__, __LINE__);
	nodes_array = blobmsg_open_array(b, "hosts");
	uci_foreach_element(&pkg->sections, e) {
		const char *macstr;
		uint8_t hwaddr[6] = {0};

		section = uci_to_section(e);
		if (strncmp(section->type, "host", 4))
			continue;

		macstr = uci_lookup_option_string(ctx, section, "macaddr");
		if (!macstr)
			continue;

		hwaddr_aton(macstr, hwaddr);

		for (i = 0; i < NODE_HTABLE_SIZE; i++) {
			if (hlist_empty(&priv->host.node_htable[i]))
				continue;

			hlist_for_each_entry(p, &priv->host.node_htable[i], hlist) {
				void *arr, *table;


				if (p->ignore)
					continue;

				if (hwaddr_is_zero(p->hwaddr) || memcmp(hwaddr, p->hwaddr, 6))  // TODO: unlikely..
					continue;
					printf("%s %d\n", __func__, __LINE__);

				table1 = blobmsg_open_table(b, NULL);
				hwaddr_ntoa(p->hwaddr, mac_str);
				blobmsg_add_string(b, "macaddr", macstr);

				blobmsg_add_string(b, "ipaddr", p->ipv4addr);
				blobmsg_add_string(b, "address_source", topology_stringify_address_types(p->addrsrc));
				blobmsg_add_u64(b, "lease_time_remaining", p->leasetmrmn);
				blobmsg_add_string(b, "device", p->device);
				blobmsg_add_string(b, "network", p->network);

				blobmsg_add_string(b, "interface_type", topology_stringify_types(p->intf_type));
				blobmsg_add_string(b, "hostname", p->hostname);

				blobmsg_add_u8(b, "active", p->active);
				tmp_t = p->active_last_change;
				info = localtime(&tmp_t);
				strftime(str_tm, sizeof(str_tm), "%Y-%m-%dT%H:%M:%SZ", info);
				blobmsg_add_string(b, "active_last_change", str_tm);
				blobmsg_add_u64(b, "active_connections", p->active_connections);
				//Here display the list of IPV4 addresses

				arr = blobmsg_open_array(b, "ipv4addr");
				for (j = 0; j < p->ipv4addr_count; j++)
					blobmsg_add_string(b, NULL, p->ipv4addr_list[j]);
				blobmsg_close_array(b, arr);
				//Here we display the list of IPV6 addresses
				arr = blobmsg_open_array(b, "ipv6addr");
				for (j = 0; j < p->ipv6addr_count; j++)
					blobmsg_add_string(b, NULL, p->ipv6addr[j]);
				blobmsg_close_array(b, arr);
				table = blobmsg_open_table(b, "stats");
				blobmsg_add_u64(b, "tx_packets", p->stats.tx_packets);
				blobmsg_add_u64(b, "tx_bytes", p->stats.tx_bytes);
				blobmsg_add_u64(b, "rx_packets", p->stats.rx_packets);
				blobmsg_add_u64(b, "rx_bytes", p->stats.rx_bytes);
				blobmsg_close_table(b, table);
				blobmsg_close_table(b, table1);
			}
		}
	}

	blobmsg_close_array(b, nodes_array);

	uci_unload(ctx, pkg);
	uci_free_context(ctx);
}

struct host_node *host_topo_node_init(struct topologyd_private *priv, uint8_t *mac_addr)
{
	struct host_node *p = NULL;
	char mac_str[18] = { 0 };
	time_t tmp_t;
	struct tm *info;
	char str_tm[20];

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!priv)
		return -1;

	if (priv->host.num_nodes >= HOST_NODE_MAX) {
		err("Number of host more than MAX limit!\n");
		return -1;
	}

	if (mac_addr != NULL) {
		int ret;

		p = host_node_lookup(priv->host.node_htable, mac_addr);
		if (!p) {
			p = host_node_add(priv->host.node_htable, mac_addr);
			if (p == NULL)
				return -1;
		}

		/* Here we need to write this info in the config file
		 */
		hwaddr_ntoa(mac_addr, mac_str);
		config_add_default_host_mac("hosts", "host", mac_str);
		host_change_active_state(priv, p, 0);

		ret = host_get_wifi_data(priv, p);
		//if (ret != 0)
		//	err("Unable to fetch interface type\n");
		if (p->intf_type == HOST_TYPE_WIFI)
			config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type", "wifi");
		else/* if (p->intf_type == HOST_TYPE_ETHER) */{
			config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type", "eth");
		}

		ret = host_get_ipaddr_hostname(priv, mac_str, p);
		if (ret != 0) {
			printf("%s  %d\n", __func__, __LINE__);
			err("Unable to fetch ip adddress and host name\n");
			p->is_ipaddr = 0;
		} else {
			printf("%s  %d\n", __func__, __LINE__);
			p->is_ipaddr = 1;
		}


	}
	return p;
}

struct host_node *host_topo_node_add(struct topologyd_private *priv, struct node *node, uint8_t *mac_addr, uint8_t is_wifi)
{
	struct host_node *p = NULL;
	char mac_str[18] = { 0 };
	time_t tmp_t;
	struct tm *info;
	char str_tm[32] = {0};
	int ret = 0, i = 0;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!priv)
		return NULL;

	if (priv->host.num_nodes >= HOST_NODE_MAX) {
		err("Number of host more than MAX limit!\n");
		return NULL;
	}

	if (node != NULL) {
		bool send_ev = false;

		if (!memcmp(node->hwaddr, priv->ieee1905_macaddr, 6)) {
			p = &priv->selfnode;
			//Here the node is the selfnode
			//self node does not get added in the hosts
			return 0;
		}

		/* Add node if not present in host.
		 *
		 * In case the node is present, update the node's values
		 * and the timestamp.
		 */
		p = host_node_lookup(priv->host.node_htable, node->hwaddr);
		if (!p) {
			p = host_node_add(priv->host.node_htable, node->hwaddr);
			if (p == NULL)
				return NULL;
			send_ev = true;
		}
		p->is_ieee1905_node = 1;
		dbg("%s %d ||||||||| setting node "MACFMT" to active\n", __func__, __LINE__, MAC2STR(p->hwaddr));

		//here we need to write  this info in the config file
		hwaddr_ntoa(node->hwaddr, mac_str);
		config_add_default_host_mac("hosts", "host", mac_str);

		// TODO review the if case
		if (p->intf_type != HOST_TYPE_WIFI || (p->intf_type == HOST_TYPE_WIFI && is_wifi))
			host_change_active_state(priv, p, 1);
		p->active_last_change = time(NULL);
		tmp_t = p->active_last_change;
		info = localtime(&tmp_t);
		strftime(str_tm, sizeof(str_tm), "%Y-%m-%dT%H:%M:%SZ", info);
		config_set_host_option("hosts", "host", "macaddr", mac_str, "active_last_change", str_tm);
		//TODO Here Adding the ipv4 address and hostname and ipv6 address
		//As these params will come in the Higher layer response it needs to be
		//populated from there

		/* This is a ieee1905 node so we get the interface type via the
		 * * topology response media type
		 * */
		//if (p->intf_type != HOST_TYPE_WIFI)
		host_get_wifi_data(priv, p);
		//if (ret != 0)
		//	err("Unable to fetch interface type\n");
	//	if (p->intf_type == HOST_TYPE_WIFI)
	//		config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type", "wifi");
	//	else/* if (p->intf_type == HOST_TYPE_ETHER) */{
	//		config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type", "eth");
	//	}

		bool has_ip = p->is_ipaddr;

		ret = host_get_ipaddr_hostname(priv, mac_str, p);
		if (ret != 0) {
			printf("%s  %d\n", __func__, __LINE__);
			err("Unable to fetch ip adddress and host name\n");
			p->is_ipaddr = 0;
		} else {
			printf("%s  %d\n", __func__, __LINE__);
			p->is_ipaddr = 1;
			if (!has_ip && p->intf_type == HOST_TYPE_WIFI && !p->is_in_assoclist) {
				if (!host_get_arping_status(p->ipv4addr)) {
					host_send_client_event(priv, p, 1);
					host_change_active_state(priv, p, 1);
				}
				else
					host_change_active_state(priv, p, 0);
			}
		}

		if (send_ev && p->active && !is_wifi && p->is_ipaddr)
			host_send_client_event(priv, p, 1);
	} else if (mac_addr != NULL) {
		bool send_ev = false;

		if (!memcmp(mac_addr, priv->ieee1905_macaddr, 6)) {
			//Here the node is the selfnode
			//self node does not get added in the hosts
			return 0;
		}

		p = host_node_lookup(priv->host.node_htable, mac_addr);
		if (!p) {
			p = host_node_add(priv->host.node_htable, mac_addr);
			if (p == NULL)
				return NULL;
			send_ev = true;
		}
		p->is_ieee1905_node = 0;
		hwaddr_ntoa(mac_addr, mac_str);
		//here we need to write this info in the config file
		config_add_default_host_mac("hosts", "host", mac_str);
		if (p->active != 1) {
			dbg("%s %d ||||||||| setting node %s to active\n", __func__, __LINE__, mac_str);
			if (p->intf_type != HOST_TYPE_WIFI || (p->intf_type == HOST_TYPE_WIFI && is_wifi))
				host_change_active_state(priv, p, 1);
			p->active_last_change = time(NULL);
			tmp_t = p->active_last_change;
			info = localtime(&tmp_t);
			strftime(str_tm, sizeof(str_tm), "%Y-%m-%dT%H:%M:%SZ", info);
			config_set_host_option("hosts", "host", "macaddr", mac_str, "active_last_change", str_tm);
			//Here Adding the ipv4 address and hostname and ipv6 address
		}

		if (is_wifi == 0 && p->intf_type != HOST_TYPE_WIFI) {
			ret = host_get_wifi_data(priv, p);
			if (ret != 0)
				err("Unable to fetch interface type\n");
		} else {
			config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type", "wifi");
			p->intf_type = HOST_TYPE_WIFI;
		}

		if (p->intf_type == HOST_TYPE_WIFI)
			config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type", "wifi");
		else/* if (p->intf_type == HOST_TYPE_ETHER) */{
			config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type", "eth");
		}


		bool has_ip = p->is_ipaddr;

		ret = host_get_ipaddr_hostname(priv, mac_str, p);
		if (ret != 0) {
			printf("%s  %d\n", __func__, __LINE__);
			err("Unable to fetch ip adddress and host name\n");
			p->is_ipaddr = 0;
		} else {
			printf("%s  %d\n", __func__, __LINE__);
			p->is_ipaddr = 1;
			if (!has_ip && p->intf_type == HOST_TYPE_WIFI && !p->is_in_assoclist) {
				if (!host_get_arping_status(p->ipv4addr)) {
					//host_send_client_event(priv, p, 1);
					host_change_active_state(priv, p, 1);
				}
				else
					host_change_active_state(priv, p, 0);
			}
		}

		if (send_ev && p->active && !is_wifi && p->is_ipaddr)
			host_send_client_event(priv, p, 1);

	}

	if (p && p->al_node)
		host_topo_node_add(priv, p->al_node, NULL, is_wifi);

	return p;
}

struct host_node *host_topo_node_del(struct topologyd_private *priv, struct node *node, uint8_t *mac_addr, uint8_t is_wifi)
{
	struct host_node *p = NULL;
	char mac_str[18] = { 0 };
	time_t tmp_t;
	struct tm *info;
	char str_tm[20];

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!priv)
		return -1;

	if (node != NULL) {
		if (!memcmp(node->hwaddr, priv->ieee1905_macaddr, 6)) {
			p = &priv->selfnode;
			//Here the node is the selfnode
			//self node does not get added in the hosts
			return NULL;
		}

		/*
		 * In case the node is present, update the node's values
		 * and the timestamp.
		 */
		p = host_node_lookup(priv->host.node_htable, node->hwaddr);
		if (p) {
			if (p->active)
				host_send_client_event(priv, p, 0);
			host_change_active_state(priv, p, 0);
			p->is_ipaddr = 0;

		} else {
			return NULL;
		}
	} else if (mac_addr != NULL) {
		p = host_node_lookup(priv->host.node_htable, mac_addr);
		if (p) {
			if (p->active != 0) {
				if (p->is_ipaddr)
					host_send_client_event(priv, p, 0);
				host_change_active_state(priv, p, 0);
			}
		} else {
			return NULL;
		}
	}
	if (p && is_wifi) {
		config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type", "wifi");
		p->intf_type = HOST_TYPE_WIFI;
	}

	if (p && p->al_node)
		host_topo_node_del(priv, p->al_node, NULL, is_wifi);

	return p;
}

int host_append_dhcpv4_info(char *mac_addr, struct host_node *p)
{
	FILE *leases;
	char line[256];
	long leasetime;
	char macaddr[24];
	char ipaddr[24];
	char hostname[64];
	char clid[256];
	int found = 0;

	dbg("Inside %s %d macis %s\n", __func__, __LINE__, mac_addr);

	if (!p)
		return -1;

	p->ipv4addr_count = 0;
	leases = fopen("/var/dhcp.leases", "r");
	if (leases != NULL) {
		while (fgets(line, sizeof(line), leases) != NULL) {
			remove_newline(line);
			if (sscanf(line, "%ld %24s %24s %64s %256s", &leasetime, macaddr, ipaddr, hostname, clid) == 5) {
				dbg("%s %d ipv4addr = [%s]\n", __func__, __LINE__, ipaddr);
				if (strcasecmp(mac_addr, macaddr) == 0) {
					if (found == 0) {
						strncpy(p->hostname, hostname, 64);
						strncpy(p->ipv4addr, ipaddr, 24);
						found = 1;
						p->leasetmrmn = leasetime - (int)time(NULL);

					}
					strncpy(p->ipv4addr_list[p->ipv4addr_count], ipaddr, 24);
					p->ipv4addr_count = (p->ipv4addr_count + 1);
				}
			}
		}
		fclose(leases);
	}
	if (found == 0)
		return 1;
	else
		return 0;
}

void host_append_dhcpv6_info(struct host_node *p)
{
	FILE *hosts6;
	char line[512];
	char hostname[64];
	char ip6addr[128];

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!p)
		return;

	p->ipv6addr_count = 0;
	hosts6 = fopen("/tmp/hosts/odhcpd", "r");
	if (hosts6 != NULL) {
		while (fgets(line, sizeof(line), hosts6) != NULL) {
			remove_newline(line);

			if (sscanf(line, "%64s %128s", ip6addr, hostname)) {
				if (strcasecmp(p->hostname, hostname) == 0) {
					dbg("ipv6 address %s %d\n", __func__, __LINE__);
					strncpy(p->ipv6addr[p->ipv6addr_count], ip6addr, 128);
					p->ipv6addr_count = p->ipv6addr_count + 1;
				}
			}
		}
		fclose(hosts6);
	}
}

void mask_num_to_bitmap(int num, uint32_t *map)
{
	int i;

	for (i = 0; i < num; i++)
		*map |= (1 << i);
}

void network_dump_parser(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct topologyd_private *p = (struct topologyd_private *)req->priv;
	struct json_object *json_msg;
	char *json_str;

	json_str = blobmsg_format_json(msg, true);
	if (!json_str)
		return;

	json_msg = json_tokener_parse(json_str);
	if (!json_msg)
		goto out_str;

	if (p->network_dump) {
		json_object_put(p->network_dump);
		p->network_dump = NULL;
	}

	if (!json_object_is_type(json_msg, json_type_object))
		goto out_json;

	p->network_dump = json_object_get(json_msg);

out_json:
	json_object_put(json_msg);
out_str:
	free(json_str);
}

void host_get_network(struct topologyd_private *priv, struct host_node *p)
{
	struct json_object *interfaces;
	char *json_str;
	int i = 0, len = 0;
	int ret = 0;

	if (!priv || !p)
		return ;

	if (!json_object_is_type(priv->network_dump, json_type_object))
		return;

	json_object_object_get_ex(priv->network_dump, "interface", &interfaces);
	if (!interfaces)
		return;

	len = json_object_array_length(interfaces);

	for (i = 0; i < len; i++) {
		struct json_object *sta_obj, *jobj_network, *ipv4_arr;
		const char *network;
		int ipv4_len, j;

		sta_obj = json_object_array_get_idx(interfaces, i);
		json_object_object_get_ex(sta_obj, "interface", &jobj_network);
		if (!jobj_network)
			continue;

		network = json_object_get_string(jobj_network);
		if (!network)
			continue;

		dbg("found network = %s\n", network);
		json_object_object_get_ex(sta_obj, "ipv4-address", &ipv4_arr);
		if (!ipv4_arr)
			continue;

		ipv4_len = json_object_array_length(ipv4_arr);
		for (j = 0; j < ipv4_len; j++) {
			struct in_addr ip, snet, host, rslt;
			struct json_object *jobj_ptr, *iface;
			char *ipaddr;
			int netmask;
			uint32_t mask = 0;

			iface = json_object_array_get_idx(ipv4_arr, j);
			if (!iface)
				continue;
			json_object_object_get_ex(iface, "mask", &jobj_ptr);
			if (!jobj_ptr)
				continue;

			netmask = json_object_get_int(jobj_ptr);

			json_object_object_get_ex(iface, "address", &jobj_ptr);
			if (!jobj_ptr)
				continue;

			ipaddr = json_object_get_string(jobj_ptr);
			if (!ipaddr)
				continue;

			mask_num_to_bitmap(netmask, &mask);

			inet_pton(AF_INET, ipaddr, &(ip.s_addr));
			inet_pton(AF_INET, p->ipv4addr, &(host.s_addr));

			snet.s_addr = (ip.s_addr & mask);
			rslt.s_addr = (host.s_addr & mask);
			if ((snet.s_addr ^ rslt.s_addr) == 0) {
				dbg("SUBNET %s MATCHED %s!!!\n", network, p->ipv4addr);
				snprintf(p->network, 16, "%s", network);
				return;
			}
		}
	}
}

static void host_get_network_dump(struct topologyd_private *priv)
{
	struct blob_buf bb = {0};
	int ret;
	uint32_t id;
	int status;

	blob_buf_init(&bb, 0);

	dbg("Calling 'interfacedump'!!\n");

	if (!priv)
		return;

	status = ubus_lookup_id(priv->ctx, "network.interface", &id);
	if (status != UBUS_STATUS_OK) {
		err("object 'interfacedump' not present!\n");
		return OBJECT_INVALID;
	}

	ret = ubus_invoke(priv->ctx, id, "dump", bb.head,
			network_dump_parser, priv, 2 * 1000);
	if (ret) {
		dbg("Failed to get network.interface dump (ret = %d)\n", ret);
		blob_buf_free(&bb);
		return -1;
	}

	blob_buf_free(&bb);
}

/* TODO: taken from questd */
static int active_connections(char *ipaddr)
{
	FILE *f;
	int i;
	char *p, line[512];
	char *saveptr;
	int connum = 0;

	f = fopen("/proc/net/nf_conntrack", "r");

	if (f != NULL) {
		while (fgets(line, sizeof(line) - 1, f)) {
			for (i = 0, p = strtok_r(line, " ", &saveptr); p; i++, p = strtok_r(NULL, " ", &saveptr)) {
				if (i == 6 && strcmp(p+4, ipaddr) == 0)
					connum++;
			}
		}

		fclose(f);
	}

	return connum;
}

static void get_dhcp_options(struct host_node *host)
{
	FILE *opts;
	char line[2048];
	char macaddr[24];
	char vcid[128];
	char clid[128];
	char ucid[128];

	/* /var/dhcp.client.options format */
	/* <MACADDR> vcid=<VENDOR_CLASS_ID> clid=<CLIENT_ID> ucid=<USER_CLASS_ID> */
	/* vcid: Option 60, clid: Option 61, ucid: Option 77 */
	opts = fopen("/var/dhcp.client.options", "r");

	if (opts != NULL) {
		while (fgets(line, sizeof(line), opts) != NULL) {
			uint8_t hwaddr[6] = {0};

			remove_newline(line);
			sscanf(line, "%24s vcid=%128s clid=%128s ucid=%128s",
					macaddr,
					vcid,
					clid,
					ucid);

			hwaddr_aton(macaddr, hwaddr);

			if (!memcmp(hwaddr, host->hwaddr, 6)) {
				if (strcmp(vcid, "-") != 0)
					strncpy(host->dhcpopts.vcid, vcid, 128);

				if (strcmp(clid, "-") != 0)
					strncpy(host->dhcpopts.clid, clid, 128);

				if (strcmp(ucid, "-") != 0)
					strncpy(host->dhcpopts.ucid, ucid, 128);

				break;
			}
		}
		fclose(opts);
	}
}

static unsigned long long get_netdev_stats(char *device, char *stat)
{
	FILE *snfile;
	char sysnet[256];
	unsigned long long val = 0;

	if (!device || !stat)
		return 0;

	snprintf(sysnet, 128, "/sys/class/net/%s/statistics/%s", device, stat);

	snfile = fopen(sysnet, "r");
	if (NULL != snfile) {
		fscanf(snfile, "%llu", &val);
		fclose(snfile);
	}

	return val;
}

void get_bridge_info(struct host_node *host, char *device)
{
	char brinfo[16];
	int aging, timer;
	char macstr[18] = {0};

	if (!host)
		return;

	hwaddr_ntoa(host->hwaddr, macstr);

	aging = 255;
	host->port_no = 0;
	if (strstr(device, "br-")) {
		chrCmd(brinfo, 16, "brctl showmacs %s | grep %s | awk '{print$1\":\"$4}'", device, macstr);
		printf("brinfo %s\n", brinfo);
		sscanf(brinfo, "%d:%d.%d", &host->port_no, &aging, &timer);
	}
}


static void get_ethernet_interface(struct host_node *host, char *bridge, int portno)
{
	DIR *dir;
	struct dirent *ent;
	FILE *pfile;
	char path_brif[256];
	char path_portno[512];
	unsigned int pno = 0;

	dbg("Inside %s %d \n", __func__, __LINE__);

	if (portno == 0 || !host)
		return;

	snprintf(path_brif, 256, "/sys/class/net/%s/brif/", bridge);

	dir = opendir(path_brif);

	if (dir != NULL) {
		while ((ent = readdir(dir)) != NULL) {
			if (ent->d_name[0] == '.')
				continue;

			snprintf(path_portno, 512, "/%s/%s/port_no", path_brif, ent->d_name);
			pfile = fopen(path_portno, "r");

			if (pfile != NULL) {
				printf("%s %d\n", __func__, __LINE__);
				fscanf(pfile, "0x%x", &pno);
				fclose(pfile);
			}

			if (pno != 0 && portno == pno) {
				strncpy(host->device, ent->d_name, 16);
				break;
			}
		}
		closedir(dir);
	}
}

int get_ifaddr(char *ifname, char *buf)
{
	int fd;
	struct ifreq ifr;
	int rv = -1;

	if (!ifname || !buf)
		return rv;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (!fd)
		return rv;

	ifr.ifr_addr.sa_family = AF_INET;

	strncpy(ifr.ifr_name, ifname, 15);
	rv = ioctl(fd, SIOCGIFADDR, &ifr);
	if (!rv)
		strncpy(buf, inet_ntoa(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr), 15);

	close(fd);

	/* display result */

	return rv;
}

int get_ifmask(char *ifname, char *buf)
{
	int fd;
	struct ifreq ifr;
	int rv = -1;

	if (!ifname || !buf)
		return rv;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (!fd)
		return rv;

	ifr.ifr_addr.sa_family = AF_INET;

	strncpy(ifr.ifr_name, ifname, 15);
	rv = ioctl(fd, SIOCGIFNETMASK, &ifr);
	if (!rv)
		strncpy(buf, inet_ntoa(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr), 15);

	close(fd);

	/* display result */

	return rv;
}

int host_get_ipaddr_hostname(struct topologyd_private *priv,
		char *macaddr_str, struct host_node *p)
{
	FILE *arpt;
	char line[256];
	int clidx = 0;
	int hw, flag;
	char mask[256];
	int i, j;
	int ret = -1;
	char macaddr[18] = { 0 };
	char ip_addr[24] = { 0 };
	char device[16] = { 0 };

	dbg("Inside %s %d mac=[%s]\n", __func__, __LINE__, macaddr_str);

	if (!macaddr_str || !p)
		return ret;

	// Walk through ARP table to get the ip address
	arpt = fopen("/proc/net/arp", "r");
	if (arpt != NULL) {
		while (fgets(line, sizeof(line), arpt) != NULL) {
			remove_newline(line);
			if (sscanf(line, "%24s 0x%d 0x%d %24s %256s %32s",
				ip_addr, &hw, &flag, macaddr, mask, device)) {
				dbg("macaddr_str %s macaddr %s, ip %s\n", macaddr_str, macaddr, ip_addr);
				// Search the  MAC address entry in the arp table
				if (strncmp(macaddr_str, macaddr, 18) == 0) {
					char mask[24] = {0}, addr[24] = {0};

					// Append DHCPv4 information: Hostname, IPv4 Address and Lease Time
					ret = host_append_dhcpv4_info(macaddr, p);
					if (ret != 0) {
						strncpy(p->ipv4addr, ip_addr, 24);
						strncpy(p->ipv4addr_list[0], ip_addr, 24);
						p->leasetmrmn = 0;
						p->addrsrc = HOST_ADDR_STATIC;
						p->ipv4addr_count = 1;
					} else {
						p->addrsrc = HOST_ADDR_DHCP;
						//get_dhcp_options(p);
					}
					dbg("device %s\n", device);
					host_get_network(priv, p);
					get_bridge_info(p, device);

					if (p->intf_type == HOST_TYPE_ETHER || !strlen(p->device)) {
						strncpy(p->device, device, 16);
						get_ethernet_interface(p, device, p->port_no);
						//if (!strncmp(p->device, "wds", 3))
						//	p->intf_type = HOST_TYPE_WIFI;
						p->stats.tx_bytes = get_netdev_stats(p->device, "rx_bytes");
						p->stats.rx_bytes = get_netdev_stats(p->device, "tx_bytes");
						p->stats.tx_packets = get_netdev_stats(p->device, "rx_packets");
						p->stats.rx_packets = get_netdev_stats(p->device, "tx_packets");
					}
					p->active_connections = active_connections(p->ipv4addr);

					dbg("%s %d ipv4addr = [%s]\n", __func__, __LINE__, p->ipv4addr);
					// Append DHCPv6 information: IPv6 Address, DUID, Lease Time
					if (strlen(p->hostname) > 1)
						host_append_dhcpv6_info(p);
					ret = 0;
					break;
				}
			}
		}
		fclose(arpt);
	}


	return ret;
}

int host_get_wifi_data(struct topologyd_private *priv, struct host_node *p)
{
	struct json_object *sta_array;
	const char *macstr;
	int i = 0, len = 0;
	int ret = 0;

	if (!p)
		return -1;

	p->is_in_assoclist = false;

	if (!json_object_is_type(priv->wifi_stations, json_type_object))
		return -1;

	json_object_object_get_ex(priv->wifi_stations, "stations", &sta_array);
	if (!sta_array)
		return -1;
	len = json_object_array_length(sta_array);

	for (i = 0; i < len; i++) {
		struct json_object *sta_obj, *sta_obj_macaddr, *jobj_ptr, *stat_obj;
		unsigned char macaddr[6] = {0};

		sta_obj = json_object_array_get_idx(sta_array, i);
		if (!sta_obj)
			continue;

		json_object_object_get_ex(sta_obj, "macaddr", &sta_obj_macaddr);
		if (!sta_obj_macaddr)
			continue;

		macstr = json_object_get_string(sta_obj_macaddr);
		if (!macstr)
			continue;

		hwaddr_aton(macstr, macaddr);
		ret = memcmp(macaddr, p->hwaddr, 6);
		if (ret)
			continue;

		json_object_object_get_ex(sta_obj, "wdev", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		strncpy(p->device, json_object_get_string(jobj_ptr), 16);
		json_object_object_get_ex(sta_obj, "frequency", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->frequency = json_object_get_int64(jobj_ptr);
		json_object_object_get_ex(sta_obj, "rssi", &jobj_ptr);
		if (!jobj_ptr)
			continue;
		p->rssi = json_object_get_int64(jobj_ptr);
		json_object_object_get_ex(sta_obj, "snr", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->snr = json_object_get_int64(jobj_ptr);
		json_object_object_get_ex(sta_obj, "idle", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->idle = json_object_get_int64(jobj_ptr);
		json_object_object_get_ex(sta_obj, "in_network", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->in_network = json_object_get_int64(jobj_ptr);

		json_object_object_get_ex(sta_obj, "stats", &stat_obj);
		if (!stat_obj)
			continue;

		json_object_object_get_ex(stat_obj, "rx_data_bytes", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->stats.rx_bytes = json_object_get_int64(jobj_ptr);
		json_object_object_get_ex(stat_obj, "tx_total_bytes", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->stats.tx_bytes = json_object_get_int64(jobj_ptr);
		json_object_object_get_ex(stat_obj, "rx_data_pkts", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->stats.rx_packets = json_object_get_int64(jobj_ptr);
		json_object_object_get_ex(stat_obj, "tx_total_pkts", &jobj_ptr);
		if (!jobj_ptr)
			continue;

		p->stats.tx_packets = json_object_get_int64(jobj_ptr);

		dbg("%s: sta: is Wifi %s\n", __func__, macstr);
		p->intf_type = HOST_TYPE_WIFI;
		host_change_active_state(priv, p, 1);
		p->is_in_assoclist = true;
		break;
	}

	return 0;
}

void host_gather_stations(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct topologyd_private *p = (struct topologyd_private *)req->priv;
	struct json_object *json_msg;
	struct json_object *sta_array;
	struct json_object *priv_stations;
	char *json_str;
	const char *macstr;
	int i = 0, len = 0, priv_len = 0;
	int ret = 0;

	json_object_object_get_ex(p->wifi_stations, "stations", &priv_stations);
	if (!priv_stations)
		return;

	priv_len = json_object_array_length(priv_stations);

	json_str = blobmsg_format_json(msg, true);
	if (!json_str)
		return;

	json_msg = json_tokener_parse(json_str);
	if (!json_msg)
		goto out_str;

	if (!json_object_is_type(json_msg, json_type_object))
		goto out_json;

	json_object_object_get_ex(json_msg, "stations", &sta_array);
	len = json_object_array_length(sta_array);

	for (i = 0; i < len; i++) {
		struct json_object *sta_obj, *sta_obj_macaddr, *jobj_ptr, *stat_obj;
		unsigned char macaddr[6] = {0};

		sta_obj = json_object_array_get_idx(sta_array, i);
		json_object_array_put_idx(priv_stations, priv_len++, json_object_get(sta_obj));
	}

out_json:
	json_object_put(json_msg);
out_str:
	free(json_str);
}

static int host_invoke_wifi_nodes(struct topologyd_private *priv,
		char *intf_name)
{
	struct blob_buf bb = {};
	int ret;
	uint32_t id;
	int status;

	blob_buf_init(&bb, 0);

	if (priv == NULL || intf_name == NULL)
		return -1;

	dbg("Calling 'assoclist' for fh-iface = %s\n", intf_name);

	status = ubus_lookup_id(priv->ctx, intf_name, &id);
	if (status != UBUS_STATUS_OK) {
		err("object '%s' not present!\n", intf_name);
		return OBJECT_INVALID;
	}

	ret = ubus_invoke(priv->ctx, id, "stations", bb.head,
			host_gather_stations, priv, 2 * 1000);
	if (ret) {
		dbg("Failed to get assoclist (ret = %d)\n", ret);
		blob_buf_free(&bb);
		return -1;
	}

	blob_buf_free(&bb);
	return 0;

}

void host_event_handler(struct ubus_context *ctx,
		struct ubus_event_handler *ev, const char *type,
		struct blob_attr *msg)
{
	int ret = 0;
	struct cmdu_cstruct *cstruct = NULL;
	uint16_t msg_type = 0;
	uint16_t msg_mid = 0;
	uint8_t *tlv = NULL;
	uint32_t len = 0;
	char itfr_name[256] = {0};
	char *str;

	struct topologyd_private *priv =
		container_of(ev, struct topologyd_private, wifi_evh);

	if (!msg)
		return;

	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	info("[ &Host = %p ] Received [event = %s]  [val = %s]\n",
		priv, type, str);
	host_wifi_sta_event_handler(priv, msg);
	free(str);
}

void host_wifi_sta_event_handler(void *c, struct blob_attr *msg)
{
	return;

	struct topologyd_private *p = (struct topologyd_private *)c;
	char ifname[16] = {0}, event[16] = {0};
	struct blob_attr *tb[3];
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	bool add = false, del = false;

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));

	if (!tb[0] || !tb[1] || !tb[2])
		return;

	strncpy(ifname, blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);

	add = !strcmp(event, "connected");
	del = !strcmp(event, "disconnected");

	if (add || del) {
		struct blob_attr *data[1];
		static const struct blobmsg_policy data_attr[1] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		};
		char mac_str[18] = {0}, bssid_str[18] = {0};

		uint8_t mac[6] = {0}, bssid[6] = {0};

		blobmsg_parse(data_attr, 1, data, blobmsg_data(tb[2]),
			blobmsg_data_len(tb[2]));

		if (!data[0])
			return;

		strncpy(mac_str, blobmsg_data(data[0]), sizeof(mac_str) - 1);

		if (!hwaddr_aton(mac_str, mac))
			return;

		if (add) {
			struct host_node *n;

			n = host_topo_node_add(p, NULL, mac, 1);
		}

		else if (del)
			host_topo_node_del(p, NULL, mac, 1);
	}
}

int host_get_interface_type(struct topologyd_private *priv)
{
	struct json_object *arr;
	struct u_list c;
	int i = 0;
	int ret = 0;
	char mac_str[18] = { 0 };
	time_t tmp_t;
	struct tm *info;
	char str_tm[20];

	if (priv->wifi_stations) {
		json_object_put(priv->wifi_stations);
		priv->wifi_stations = NULL;
	}

	priv->wifi_stations = json_object_new_object();
	if (!priv->wifi_stations)
		return -1;

	arr = json_object_new_array();
	if (!arr)
		return -1;

	json_object_object_add(priv->wifi_stations, "stations", arr);
	host_get_wifi_nodes(priv->ctx, &c);
	for (i = 0 ; i < c.count ; i++) {
		ret = host_invoke_wifi_nodes(priv, c.ifname[i]);
		if (ret != 0) {
			err("Unable to get the assoclist\n");
			return -1;
		}
	}

	return ret;
}

/*Here the host periodic refresh timer does as below:
 * 1. run the topology refresh algo such that all the new
 *    non -1905 neighbors can be fetched
 * 2. Check all the nodes of type ETHER and get there status
 *    from the arp table.
 */
void host_periodic_refresh(struct uloop_timeout *t)
{
	struct topologyd_private *priv =
		container_of(t, struct topologyd_private, host_refresh_timer);

	dbg("Refresh host topology...\n");

	/*here we send the topology query to selfnode each*/
	/*time the algo runs to keep self node updated */
	topologyd_send_ieee1905_topology_query(priv, NULL);
	host_run_status_check(priv);
	uloop_timeout_set(&priv->host_refresh_timer, HOST_REFRESH_TIMER * 1000);
}

int host_get_arping_status(char *ipaddr_str)
{
	char cmd[256] = {0};
	FILE *fp = NULL;
	char *line = NULL;
	size_t len = 0;
	int found = 0;
	int ret = -1;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (ipaddr_str == NULL)
		return -1;

	snprintf(cmd, 256, ARPING_CMD, ipaddr_str);

	fp = popen(cmd, "r");
	if (fp == NULL) {
		err("unable to execute command");
		return -1;
	}

	dbg("Inside %s %d command is %s\n", __func__, __LINE__, cmd);
	while (getline(&line, &len, fp) != -1) {
		if (len < 2) {
			free(line);
			len = 0;
			continue;
		}
		ret = strncmp(line, "1", 1);
		break;
	}
	free(line);
	pclose(fp);
	return ret;
}

int host_get_neigh_status(uint8_t *mac_addr)
{
	char cmd[256] = {0};
	FILE *fp = NULL;
	char *line = NULL;
	size_t len = 0;
	char macaddr_str[18];
	int found = 0;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (mac_addr == NULL)
		return NULL;

	hwaddr_ntoa(mac_addr, macaddr_str);
	snprintf(cmd, 256, NEIGH_CMD, macaddr_str);

	fp = popen(cmd, "r");
	if (fp == NULL) {
		err("unable to execute command");
		return -1;
	}

	dbg("Inside %s %d command is %s\n", __func__, __LINE__, cmd);
	while (getline(&line, &len, fp) != -1) {
		if (len < 17) {
			free(line);
			len = 0;
			continue;
		}
		found = 1;
		break;
	}

	free(line);
	pclose(fp);

	if (found == 0)
		return -1;
	else
		return 0;
}

int host_check_ip_addr(struct host_node *p)
{

	char mac_str[18] = { 0 };
	int ret = -1, i = 0;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (p == NULL)
		return -1;

	hwaddr_ntoa(p->hwaddr, mac_str);
	for (i = 0; i < p->ipv4addr_count; i++)
		config_del_host_list("hosts", "host", "macaddr", mac_str,
			"ipv4address", p->ipv4addr_list[i]);
	for (i = 0; i < p->ipv6addr_count; i++)
		config_del_host_list("hosts", "host", "macaddr", mac_str,
			"ipv6address", p->ipv6addr[i]);
	//ret = host_get_ipaddr_hostname(priv, mac_str, p);
	if (ret != 0)
		err("Unable to fetch ip adddress and host name\n");
	else {
		p->is_ipaddr = 1;
		config_set_host_option("hosts", "host", "macaddr", mac_str,
				"ipaddress", p->ipv4addr);
		config_set_host_option("hosts", "host", "macaddr", mac_str,
				"hostname", p->hostname);

		for (i = 0; i < p->ipv4addr_count; i++)
			config_set_host_list("hosts", "host", "macaddr", mac_str,
				"ipv4address", p->ipv4addr_list[i]);
		for (i = 0; i < p->ipv6addr_count; i++)
			config_set_host_list("hosts", "host", "macaddr",
				mac_str, "ipv6address", p->ipv6addr[i]);
	}
	return ret;
}

/*
{ "topology.host": {"action":"connect","macaddr":"fe:ce:90:d7:5e:64","ipaddr":"192.168.1.157","network":"lan"} }
{ "topology.host": {"action":"disconnect","macaddr":"fe:ce:90:d7:5e:64","ipaddr":"192.168.1.157","network":"lan"} }
*/

void host_send_client_event(struct topologyd_private *priv, struct host_node *p, uint8_t state)
{
	char mac[18] = {0}, src_mac[18] = {0};
	struct blob_buf bb = {0};

	if (!p || !priv)
		return;

	dbg("Inside %s...\n", __func__);

	blob_buf_init(&bb, 0);
	hwaddr_ntoa(p->hwaddr, mac);
	blobmsg_add_u8(&bb, "active", state);
	blobmsg_add_string(&bb, "macaddr", mac);
	blobmsg_add_string(&bb, "ipaddr", p->ipv4addr);
	blobmsg_add_string(&bb, "network", p->network);

	ubus_send_event(priv->ctx, "topology.host", bb.head);

	blob_buf_free(&bb);
}

void host_change_active_state(struct topologyd_private *priv, struct host_node *p, uint8_t state)
{
	char mac_str[18] = { 0 };
	time_t tmp_t;
	struct tm *info;
	char str_tm[32] = { 0 };

	dbg("Inside %s %d\n", __func__, __LINE__);
	hwaddr_ntoa(p->hwaddr, mac_str);
	p->active = state;
	p->active_last_change = time(NULL);
	//here we need to write this info in the config file
	config_add_default_host_mac("hosts", "host", mac_str);
	if (state == 0x0)
		config_set_host_option("hosts", "host", "macaddr", mac_str, "active", "0");
	else {
		config_set_host_option("hosts", "host", "macaddr", mac_str, "active", "1");
	}
	tmp_t = p->active_last_change;
	info = localtime(&tmp_t);
	strftime(str_tm, sizeof(str_tm), "%Y-%m-%dT%H:%M:%SZ", info);
	config_set_host_option("hosts", "host", "macaddr", mac_str, "active_last_change", str_tm);
}

/*Here the below are checked for each host node in case of periodic run
 * 1. If Ipaddress is fetched from dhcp lease
 * 2. If node is eth then check reachable
 * 3. If node is eth and not reachable do arping
 */
void host_run_status_check(struct topologyd_private *priv)
{
	struct host_node *p;
	int i = 0, ret = -1;
	char mac_str[18] = { 0 };
	time_t tmp_t;
	struct tm *info;
	char str_tm[20];

	dbg("Inside %s %d\n", __func__, __LINE__);
	if (!priv)
		return;
	host_get_interface_type(priv);
	host_get_network_dump(priv);

	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&priv->host.node_htable[i]))
			continue;

		hlist_for_each_entry(p, &priv->host.node_htable[i], hlist) {
			bool has_ip = p->is_ipaddr;

			if (p->al_node)
				continue;

			if (hwaddr_is_zero(p->hwaddr))  // TODO: unlikely..
				continue;

			hwaddr_ntoa(p->hwaddr, mac_str);
			/*Here we check if the node has received the
			 * ip adddress
			 */
			//if (p->is_ipaddr == 0) {
			//	dbg("node %s does not have an ip address!\n", mac_str);
			//	ret = host_get_ipaddr_hostname(priv, mac_str, p);
			//	if (ret != 0)
			//		err("Unable to fetch ip adddress and host name\n");
			//	else
			//		p->is_ipaddr = true;
			//}
			/*Here we check the status of the eth node in
			 * the ip neigh
			 */
			if (p->intf_type != HOST_TYPE_ETHER && !p->is_ieee1905_node) {
				memset(p->device, 0, 16);
				host_get_wifi_data(priv, p);
				dbg("%s %d hasip = %d, %s\n", __func__, __LINE__, has_ip, p->ipv4addr);

				ret = host_get_ipaddr_hostname(priv, mac_str, p);
				if (ret != 0) {
					err("Unable to fetch ip adddress and host name\n");
					p->is_ipaddr = 0;
				} else {
					p->is_ipaddr = 1;
					if (!has_ip) {
						if (p->is_in_assoclist) {
							host_send_client_event(priv, p, 1);
						} else if (p->intf_type == HOST_TYPE_WIFI && !p->is_in_assoclist) {
							if (!host_get_arping_status(p->ipv4addr)) {
								host_send_client_event(priv, p, 1);
								host_change_active_state(priv, p, 1);
							}
							else {
								//host_send_client_event(priv, p, 0);
								host_change_active_state(priv, p, 0);
							}
						}
					}
				}
				dbg("%s %d hasip = %d, %s\n", __func__, __LINE__, has_ip, p->ipv4addr);
				continue;
			} else if (p->is_ieee1905_node)
				memset(p->device, 0, 16);


			if (host_get_ipaddr_hostname(priv, mac_str, p))
				p->is_ipaddr = 0;
			else
				p->is_ipaddr = 1;

			ret = host_get_neigh_status(p->hwaddr);
			if (ret == 0 && has_ip) {
				/*Here the node is rechable so
				 * active status is 1
				 */
				dbg("ETH node is reachable " MACFMT "\n", MAC2STR(p->hwaddr));
				if (p->active != 1) {
					//Here as the node is reachable
					//we are adding in map-topology
					host_send_client_event(priv, p, 1);
					host_change_active_state(priv, p, 1);
				}
			} else {
				/*Here as the status is not reachable
				 * we need to do arping to the node
				 * to get the status and turn unreachable accordingly
				 */
				char ipv4addr[24] = { 0 };

				if (strncmp(p->ipv4addr, ipv4addr, 24)) {
					ret = host_get_arping_status(p->ipv4addr);
					if (ret != 0) {
						dbg("ETH node is unreachable " MACFMT "\n", MAC2STR(p->hwaddr));
						if (p->active != 0) {
							host_send_client_event(priv, p, 0);
							host_change_active_state(priv, p, 0);
						}
					} else {
						if (p->active != 1) {
							dbg("ETH node is reachable " MACFMT "\n", MAC2STR(p->hwaddr));
							host_send_client_event(priv, p, 1);
							host_change_active_state(priv, p, 1);
						} else if (!has_ip)
							host_send_client_event(priv, p, 1);

					}
				} else {
					dbg("ETH node is unreachable " MACFMT "\n", MAC2STR(p->hwaddr));
					if (p->active != 0) {
						host_send_client_event(priv, p, 0);
						host_change_active_state(priv, p, 0);
					}
				}
			}
		}
	}
}

void host_process_topo_notification(uint8_t *tlv, struct topologyd_private *priv)
{
	int i = 0, j = 0;
	uint8_t *ptr;
	uint8_t *p, *ret;
	uint8_t temp = 0;
	uint16_t tlv_length = 0;

	if (tlv == NULL || priv == NULL)
		return;

	dbg("Inside topologyd process notification tlv\n");

	switch (*tlv) {
	case MAP_TLV_CLIENT_ASSOCIATION_EVENT:
	{
		uint8_t hwaddr[6] = {0};
		uint8_t assoc_event = 0;

		memcpy(&assoc_event, (tlv + 1 + 2 + 6 + 6), 1);
		memcpy(hwaddr, tlv + 3, 6);
		dbg("Topology Notification TLV client addr " MACFMT " assoc_event 0x%02x\n",
			MAC2STR(hwaddr), assoc_event);
		if (assoc_event == 0x80) {
			struct host_node *n;

			n = host_topo_node_add(priv, NULL, hwaddr, 1);
			if (n->is_ipaddr)
				host_send_client_event(priv, n, 1);
		} else
			host_topo_node_del(priv, NULL, hwaddr, 1);

	}
	break;
	default:
		dbg(" received a default tlv\n");
	break;
	}
	i++;
}

void host_agent_event_handler(struct ubus_context *ctx,
		struct ubus_event_handler *ev, const char *type,
		struct blob_attr *msg)
{
	int ret = 0;
	uint16_t msg_type = 0;
	uint16_t msg_mid = 0;
	uint8_t *tlv = NULL;
	uint32_t len = 0;
	char itfr_name[256] = {0};

	struct topologyd_private *priv =
		container_of(ev, struct topologyd_private, agent_evh);

	if (!msg)
		return;

	//Calling the function to get the args from json msg
	ret = ieee1905_msg_from_ubus(msg, &msg_type, &msg_mid, &tlv, &len, itfr_name);
	if (ret != 0) {
		err("ieee1905_msg_from_ubus error\n");
		if (tlv != NULL)
			free(tlv);
		return;
	}
	dbg("Inside host agent event handler %hu %hu  %s", msg_type, msg_mid, itfr_name);

	if (!(msg_type == CMDU_TYPE_TOPOLOGY_NOTIFICATION)) {
		if (tlv != NULL)
			free(tlv);
		return;
	}

	dbg("After Event handler\n");

	switch (msg_type) {
	case CMDU_TYPE_TOPOLOGY_NOTIFICATION:
	{
		/* Here if the topology notification is from map-agent
		 * to notify about the wifi nodes added we need
		 * to parse the message and add the node
		 * */
		host_process_topo_notification(tlv, priv);
	}
	break;
	default:
		dbg("Event handler received default message\n");
	break;
	}

	if (tlv != NULL)
		free(tlv);
}

void host_set_interface_type(struct host_node *p)
{
	char mac_str[18] = { 0x0 };

	if (p == NULL)
		return;

	hwaddr_ntoa(p->hwaddr, mac_str);
	if (p->intf_type == HOST_TYPE_WIFI)
		config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type", "wifi");
	else
		config_set_host_option("hosts", "host", "macaddr", mac_str, "interface_type", "eth");
	config_set_host_option("hosts", "host", "macaddr", mac_str, "active", "1");

}
