/*
 * host_config.c - host configuration file
 *
 * Copyright (C) 2021 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <uci.h>

#include "debug.h"
#include "host.h"

struct uci_section *config_get_section(struct uci_context *ctx,
		struct uci_package *pkg, const char *type, const char *key,
		const char *value)
{
	struct uci_element *e;
	struct uci_section *section;

	dbg("Inside %s %d\n", __func__, __LINE__);

	/* get the wet iface section */
	uci_foreach_element(&pkg->sections, e) {
		const char *c_value;

		section = uci_to_section(e);
		if (strcmp(section->type, type))
			continue;

		c_value = uci_lookup_option_string(ctx, section, key);
		if (c_value && !strcmp(c_value, value))
			return section;
	}

	return NULL;
}

struct uci_section *config_add_section(struct uci_context *ctx,
		struct uci_package *pkg, const char *config, const char *type,
		char *name, const char *key, const char *value)
{
	struct uci_section *section = NULL;
	struct uci_ptr ptr = {0};
	int rv = -1, ret = 0;
	char name_section[50] = { 0 };

	dbg("Inside %s %d\n", __func__, __LINE__);

	static const char s[2] = ":";
	char *token;
	char mac_name[18] = { 0 };

	strcpy(mac_name, value);
	/* get the first token */
	token = strtok(mac_name, s);

	/* walk through other tokens */
	while (token != NULL) {
		strcat(name_section, token);
		token = strtok(NULL, s);
	}

	section = config_get_section(ctx, pkg, type, key, value);
	if (!section) {
		rv = uci_add_section(ctx, pkg, type, &section);
		if (rv)
			goto out_pkg;

		ptr.p = pkg;
		ptr.s = section;
		ptr.value = name_section;
		ret = uci_rename(ctx, &ptr);

		rv = uci_save(ctx, pkg);
		if (rv)
			goto out_pkg;
	}

	ptr.value = value;
	ptr.package = config;
	ptr.section = section->e.name;
	ptr.option = key;
	ptr.target = UCI_TYPE_OPTION;

	uci_lookup_ptr(ctx, &ptr, NULL, false);
	uci_set(ctx, &ptr);
	uci_save(ctx, ptr.p);

out_pkg:
	return section;
}


int config_add_default_host_mac(const char *config, const char *type,
		const char *macaddr)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_section *section;
	int rv = -1;

	dbg("Inside %s %d\n", __func__, __LINE__);

	ctx = uci_alloc_context();
	if (!ctx)
		goto out;

	if (uci_load(ctx, config, &pkg) != UCI_OK) {
		err("config file 'host' not found!\n");
		goto out_uci;
	}

	section = config_add_section(ctx, pkg, config, type, "mac_addr", "macaddr", macaddr);
	if (!section)
		return -1;

	uci_commit(ctx, &pkg, false);
out_pkg:
	uci_unload(ctx, pkg);
out_uci:
	uci_free_context(ctx);
out:
	return 0;
}

bool config_set_host_option(char *package_name,
		char *section_type, char *search_key, char *search_val,
		char *option, char *value)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!package_name || !search_val || !option || !value)
		return false;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (uci_load(ctx, package_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, section_type)) {
			struct uci_option *opt = uci_lookup_option(ctx, s,
							search_key);

			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;
			if (strcmp(opt->v.string, search_val) == 0) {
				struct uci_ptr ptr = {0};

				ptr.value = value;
				ptr.package = package_name;
				ptr.section = s->e.name;
				ptr.option = option;
				ptr.target = UCI_TYPE_OPTION;
				if (uci_lookup_ptr(ctx, &ptr, NULL, false) ||
						!UCI_LOOKUP_COMPLETE)
					break;
				if (uci_set(ctx, &ptr) == UCI_OK)
					uci_save(ctx, ptr.p);
				break;
			}
		}
	}
	uci_commit(ctx, &pkg, false);
	uci_unload(ctx, pkg);
	uci_free_context(ctx);
	return false;
}

bool config_del_host_list(char *package_name,
		char *section_type, char *search_key, char *search_val,
		char *option, char *value)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e, *x;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!package_name || !search_val || !option || !value)
		return false;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (uci_load(ctx, package_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, section_type)) {
			struct uci_option *opt = uci_lookup_option(ctx, s,
							search_key);
			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;
			if (strcmp(opt->v.string, search_val) == 0) {
				struct uci_ptr ptr = {0};

				ptr.value = value;
				ptr.package = package_name;
				ptr.section = s->e.name;
				ptr.option = option;
				ptr.target = UCI_TYPE_OPTION;
				if (uci_del_list(ctx, &ptr) == UCI_OK)
					uci_save(ctx, ptr.p);
				break;
			}
		}
	}
	uci_commit(ctx, &pkg, false);
	uci_unload(ctx, pkg);
	uci_free_context(ctx);
	return false;
}

bool config_set_host_list(char *package_name,
		char *section_type, char *search_key, char *search_val,
		char *option, char *value)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!package_name || !search_val || !option || !value)
		return false;

	config_del_host_list(package_name, section_type, search_key,
			search_val, option, value);

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (uci_load(ctx, package_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, section_type)) {
			struct uci_option *opt = uci_lookup_option(ctx, s,
							search_key);
			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;
			if (strcmp(opt->v.string, search_val) == 0) {
				struct uci_ptr ptr = {0};

				ptr.value = value;
				ptr.package = package_name;
				ptr.section = s->e.name;
				ptr.option = option;
				ptr.target = UCI_TYPE_OPTION;
				if (uci_add_list(ctx, &ptr) == UCI_OK)
					uci_save(ctx, ptr.p);
				break;
			}
		}
	}
	uci_commit(ctx, &pkg, false);
	uci_unload(ctx, pkg);
	uci_free_context(ctx);
	return false;
}

bool config_init_host_data(char *package_name, char *section, struct topologyd_private *priv)
{

	bool ret;
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	char mac_address[18];
	uint8_t mac[6];
	char intf_type[20];

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (!package_name)
		return false;

	ctx = uci_alloc_context();
	if (!ctx)
		return false;

	if (uci_load(ctx, package_name, &pkg)) {
		uci_free_context(ctx);
		return false;
	}

	ret = false;
	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, section)) {
			struct host_node *p = NULL;
			struct uci_option *opt = uci_lookup_option(ctx, s,
							"macaddr");
			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;
			strncpy(mac_address, opt->v.string, 18);
			hwaddr_aton(mac_address, mac);

			opt = uci_lookup_option(ctx, s, "interface_type");
			if (!opt || opt->type != UCI_TYPE_STRING)
				continue;
			strncpy(intf_type, opt->v.string, 20);
			if (!strcmp(intf_type, "eth")) {
				dbg("%s IS AN ETHERNET CLIENT ACCORDING TO CONFIG\n", mac_address);
				p = host_topo_node_add(priv, NULL, mac, 0);
			} else {
				dbg("%s IS AN WIFI CLIENT ACCORDING TO CONFIG\n", mac_address);
				p = host_topo_node_add(priv, NULL, mac, 1);
			}
			if (p) {
				if (p->intf_type == HOST_TYPE_WIFI) {
					if (p->ipv4addr && host_get_arping_status(p->ipv4addr)) {
						dbg("%s %d client %s DID NOT answer arping, set as inactive\n", __func__, __LINE__, p->ipv4addr);
						host_change_active_state(priv, p, 0);
					} else {
						dbg("%s %d client %s DID answer arping, set as active\n", __func__, __LINE__, p->ipv4addr);
						host_change_active_state(priv, p, 1);
					}
				}
			}
		}
	}
	uci_unload(ctx, pkg);
	uci_free_context(ctx);
	return ret;
}

struct uci_package *config_init_package(struct uci_context **ctx, const char *config)
{
	struct uci_package *p = NULL;

	dbg("Inside %s %d\n", __func__, __LINE__);

	if (*ctx == NULL) {
		*ctx = uci_alloc_context();
	} else {
		p = uci_lookup_package(*ctx, config);
		if (p)
			uci_unload(*ctx, p);
	}

	if (uci_load(*ctx, config, &p))
		return NULL;

	return p;
}

void config_free_uci_context(struct uci_context **ctx)
{
	dbg("Inside %s %d\n", __func__, __LINE__);
	if (*ctx)
		uci_free_context(*ctx);
	*ctx = NULL;
}

